import socket
import selectors
import types
import subprocess
import struct
import fcntl


class MVMAgent:
    def __init__(self, selector: selectors.DefaultSelector):
        '''
        Initialize the mvm agent to be in a listening state
        '''
        self.port = 1234
        self.cid = self.get_vsock_cid()
        self.command = None
        self.sel = selector

    def get_vsock_cid(self) -> int:
        with open("/dev/vsock", "rb") as fd:
            r = fcntl.ioctl(fd, socket.IOCTL_VM_SOCKETS_GET_LOCAL_CID, "    ")
            cid = struct.unpack("I", r)[0]
            return int(cid)

    def start_connection(self):
        lsock = socket.socket(socket.AF_VSOCK, socket.SOCK_STREAM)
        lsock.bind((self.cid, self.port))
        lsock.listen()
        print(f"Listening on {(self.cid, self.port)}")
        lsock.setblocking(False)
        self.sel.register(lsock, selectors.EVENT_READ, data=None)

    def accept_wrapper(self, sock) -> None:
        conn, addr = sock.accept()  # Should be ready to read
        print(f"Accepted connection from {addr}")
        conn.setblocking(False)
        data = types.SimpleNamespace(addr=addr, inb=b"", outb=b"")
        events = selectors.EVENT_READ | selectors.EVENT_WRITE
        self.sel.register(conn, events, data=data)

    def service_connection(self, key, mask) -> None:
        sock = key.fileobj
        data = key.data
        if mask & selectors.EVENT_READ:
            recv_data = sock.recv(1024)  # Should be ready to read
            if recv_data:
                # Run the received command in the guest
                command = recv_data.decode('utf-8')
                proc = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
                result = proc.stdout or proc.stderr or 'OK!!'
                data.outb = result.encode('utf-8')
            else:
                print(f"Closing connection to {data.addr}")
                self.sel.unregister(sock)
                sock.close()
        if mask & selectors.EVENT_WRITE:
            if data.outb:
                print(f"Executing command received from {data.addr}")
                sent = sock.send(data.outb)  # Should be ready to write
                data.outb = data.outb[sent:]


def main():
    sel = selectors.DefaultSelector()
    mvm_agent = MVMAgent(sel)
    mvm_agent.start_connection()

    try:
        while True:
            events = sel.select(timeout=None)
            for key, mask in events:
                if key.data is None:
                    mvm_agent.accept_wrapper(key.fileobj)
                else:
                    mvm_agent.service_connection(key, mask)
    except KeyboardInterrupt:
        sel.close()


if __name__ == '__main__':
    main()
